/* eslint-disable no-param-reassign */

'use strict';

const fs = require('fs');
// const path = require('path');
const crypto = require('crypto');

const fastGlob = require('fast-glob');
const cacache = require('cacache');
const arrayDiff = require('arr-diff');

async function main({
  include,
  exclude = ['**/node_modules'],
  handler = () => {},
  alwaysHandler = () => {},
  alwaysHandlerOnly = false,
  forceUpdateCache = true,
  cachePath = './.cache/rdj',
  fastGlobOptions = {},
}) {
  const files = await Promise.all(
    (
      await fastGlob(include, {
        ...fastGlobOptions,
        unique: true,
        absolute: true,
        objectMode: true,
        ignore: exclude,
      })
    ).map(async (file) => {
      file.contents = await fs.promises.readFile(file.path);
      return file;
    }),
  );

  const cached = await cacache.ls(cachePath);
  const cacheFiles = Object.keys(cached);

  const missingFilesInCache = arrayDiff(
    files.map((x) => x.path),
    cacheFiles,
  );

  // source files are missing in cache, add them
  if (missingFilesInCache.length > 0) {
    // console.log('has missing', missingFilesInCache);
    await Promise.all(
      missingFilesInCache
        .reduce(
          (acc, fp) =>
            acc.concat(files.find((x) => x.path === fp)).filter(Boolean),
          [],
        )
        .map(async ({ path: filepath, contents: cont }) => {
          const res = await cacache.put(cachePath, filepath, cont.toString());
          const ctx = createContext({
            filepath,
            contents: cont,
            size: cont.length,
            integrity: res.sha512[0].source,
            cachePath,
            cache: null,
          });

          await alwaysHandler(ctx);
          if (alwaysHandlerOnly !== true) {
            await handler(ctx);
          }

          return ctx;
        }),
    );
    // console.log('they are written to cache');
  }

  if (cacheFiles.length > 0) {
    // console.log('cacheFiles');
    await Promise.all(
      cacheFiles.map(async (fp) => {
        const cacheFile = cached[fp];
        // most probably the source file was deleted
        if (!fs.existsSync(fp)) {
          // console.log('delete from cache:', fp);

          // ? should we delete the index too (doesn't seem to work anyway)?
          // await cacache.rm.entry(cachePath, fp);
          await cacache.rm.content(cachePath, cacheFile.integrity);
          await cacache.verify(cachePath);
          return;
        }
        const contents = await fs.promises.readFile(fp);
        const integrity = integrityFromContents(contents);

        const hash = await cacache.get.hasContent(cachePath, integrity);

        const ctx = createContext({
          filepath: fp,
          contents,
          integrity,
          size: contents.length,
          cachePath,
          cacache,
          updateCacheFile,

          // cannot have when not in cache (in `missingFilesInCache` check)
          cache: cacheFile,
          hash,
          invalid: Boolean(hash),
        });

        await alwaysHandler(ctx);

        // invalid integrity or size, do your job.
        if (!hash && alwaysHandlerOnly !== true) {
          console.log('invalid integrity or size:', fp);

          await handler(ctx);

          // if (forceUpdateCache === true) {
          //   await updateCache({ cachePath, filepath: fp, contents: cont });
          // }
        }
      }),
    );
  }
}

function integrityFromContents(contents) {
  const integrity = crypto
    .createHash('sha512')
    .update(contents)
    .digest('base64');

  return `sha512-${integrity}`;
}
function contentsFromIntegrity(integrity) {}

function createContext(ctx) {
  Object.defineProperty(ctx, 'filePath', { value: ctx.filepath });
  Object.defineProperty(ctx, 'cachepath', { value: ctx.cachePath });
  return ctx;
}

async function updateCacheFile({ cachePath, filepath, contents }) {
  const res = await cacache.put(cachePath, filepath, contents);

  return createContext({
    filepath,
    filePath: filepath,
    contents,
    size: contents.length,
    integrity: res.sha512[0].source,
  });
}

async function runner(context) {
  console.log('ctx', context);
  // updateCache(context);;;;;;;;;
}

main({
  include: ['**/*.js', '*.json'],
  alwaysHandler: runner,
  alwaysHandlerOnly: true,
}).then(() => {
  console.log('Done!');
});
